package com.flyfish.framework.logging.advice;

import com.flyfish.framework.logging.service.LogContext;
import com.flyfish.framework.logging.service.LogManager;
import lombok.RequiredArgsConstructor;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Pointcut;
import org.reactivestreams.Publisher;
import org.reactivestreams.Subscriber;
import org.reactivestreams.Subscription;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

/**
 * 日志切面
 *
 * @author wangyu
 */
@Aspect
@RequiredArgsConstructor
public class LogAdvice {

    private final LogManager logManager;

    /**
     * 切入注解
     */
    @Pointcut("execution(public * com.flyfish..*.*Controller.*(..))")
    public void pointCut() {

    }


    /**
     * 环绕切面，获取参数
     *
     * @param joinPoint 切入点
     * @return 结果
     */
    @Around("pointCut()")
    public Object process(ProceedingJoinPoint joinPoint) throws Throwable {
        // 构建上下文
        LogContext context = LogContext.of(joinPoint);
        // 结果信息
        Object result = null;
        // 是否是publisher
        boolean reactive = false;
        try {
            // 执行方法
            result = joinPoint.proceed(joinPoint.getArgs());
            // 判断是否异步
            reactive = result instanceof Publisher;
            // 异步需特殊处理，否则直接返回
            return reactive ? handleResult(result, context) : result;
        } catch (Throwable throwable) {
            context.setError(throwable);
            throw throwable;
        } finally {
            if (!reactive && context.isValid()) {
                context.setResult(result);
                logManager.tryLog(context.end());
            }
        }
    }


    /**
     * 处理结果，支持webflux
     *
     * @param result  结果
     * @param context 上下文
     */
    private Object handleResult(Object result, LogContext context) {
        Runnable runnable = () -> {
            if (context.isValid()) {
                logManager.tryLog(context.end());
            }
        };
        if (result instanceof Mono) {
            return ((Mono<?>) result)
                    .doOnSuccess(context::setResult)
                    .doOnError(context::setError)
                    .doOnTerminate(runnable);
        } else if (result instanceof Flux) {
            return ((Flux<?>) result)
                    .collectList()
                    .doOnSuccess(context::setResult)
                    .doOnError(context::setError)
                    .doOnTerminate(runnable)
                    .flatMapIterable(list -> list);
        } else {
            // 未知，不监听
            return result;
        }
    }

    @RequiredArgsConstructor
    @SuppressWarnings("all")
    private static class LogSubscriber implements Subscriber<Object> {

        private final LogContext context;

        @Override
        public void onSubscribe(Subscription s) {

        }

        @Override
        public void onNext(Object o) {
            context.setResult(o);
        }

        @Override
        public void onError(Throwable t) {
            context.setError(t);
        }

        @Override
        public void onComplete() {
            context.end();
        }
    }
}
